from __future__ import annotations

import builtins
import math
import textwrap
import unittest.mock

import pandas as pd
import pandas.testing
import polars as pl
import polars.testing
import pyarrow as pa
import pytest

import tea_tasting.utils


def test_check_scalar_typ():
    assert tea_tasting.utils.check_scalar(1, typ=int) == 1
    with pytest.raises(TypeError):
        tea_tasting.utils.check_scalar(1, typ=str)

def test_check_scalar_ge():
    assert tea_tasting.utils.check_scalar(1, ge=1) == 1
    with pytest.raises(ValueError, match="must be >="):
        tea_tasting.utils.check_scalar(1, ge=2)

def test_check_scalar_gt():
    assert tea_tasting.utils.check_scalar(1, gt=0) == 1
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.check_scalar(1, gt=1)

def test_check_scalar_le():
    assert tea_tasting.utils.check_scalar(1, le=1) == 1
    with pytest.raises(ValueError, match="must be <="):
        tea_tasting.utils.check_scalar(1, le=0)

def test_check_scalar_lt():
    assert tea_tasting.utils.check_scalar(1, lt=2) == 1
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.check_scalar(1, lt=1)

def test_check_scalar_ne():
    assert tea_tasting.utils.check_scalar(1, ne=2) == 1
    with pytest.raises(ValueError, match="must be !="):
        tea_tasting.utils.check_scalar(1, ne=1)

def test_check_scalar_is_in():
    assert tea_tasting.utils.check_scalar(1, in_={0, 1}) == 1
    with pytest.raises(ValueError, match="must be in"):
        tea_tasting.utils.check_scalar(1, in_={0, 2})


def test_auto_check_alpha():
    assert tea_tasting.utils.auto_check(0.05, "alpha") == 0.05
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "alpha")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "alpha")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "alpha")

def test_auto_check_alternative():
    assert tea_tasting.utils.auto_check("two-sided", "alternative") == "two-sided"
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(2, "alternative")
    with pytest.raises(ValueError, match="must be in"):
        tea_tasting.utils.auto_check("2s", "alternative")

def test_auto_check_confidence_level():
    assert tea_tasting.utils.auto_check(0.95, "confidence_level") == 0.95
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "confidence_level")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "confidence_level")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "confidence_level")

def test_auto_check_correction():
    assert tea_tasting.utils.auto_check(True, "correction") is True
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "correction")

def test_auto_check_equal_var():
    assert tea_tasting.utils.auto_check(True, "equal_var") is True
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "equal_var")

def test_auto_check_n_obs():
    assert tea_tasting.utils.auto_check(2, "n_obs") == 2
    assert tea_tasting.utils.auto_check((2, 3), "n_obs") == (2, 3)
    assert tea_tasting.utils.auto_check(None, "n_obs") is None
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0.5, "n_obs")
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check((0.5, 2), "n_obs")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(1, "n_obs")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check((1, 2), "n_obs")

def test_auto_check_n_resamples():
    assert tea_tasting.utils.auto_check(1, "n_resamples") == 1
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0, "n_resamples")
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0.5, "n_resamples")

def test_auto_check_power():
    assert tea_tasting.utils.auto_check(0.8, "power") == 0.8
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "power")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "power")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "power")

def test_auto_check_ratio():
    assert tea_tasting.utils.auto_check(1.5, "ratio") == 1.5
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check("str", "ratio")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "ratio")

def test_auto_check_use_t():
    assert tea_tasting.utils.auto_check(False, "use_t") is False
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "use_t")


def test_format_num():
    assert tea_tasting.utils.format_num(1.2345) == "1.23"
    assert tea_tasting.utils.format_num(0.9999) == "1.00"
    assert tea_tasting.utils.format_num(1.2345, sig=2) == "1.2"
    assert tea_tasting.utils.format_num(0.12345, pct=True) == "12.3%"
    assert tea_tasting.utils.format_num(None) == "-"
    assert tea_tasting.utils.format_num(float("nan")) == "-"
    assert tea_tasting.utils.format_num(float("inf")) == "∞"
    assert tea_tasting.utils.format_num(float("-inf")) == "-∞"
    assert tea_tasting.utils.format_num(0.00012345) == "1.23e-04"
    assert tea_tasting.utils.format_num(0.00099999) == "1.00e-03"
    assert tea_tasting.utils.format_num(9_999_999) == "9999999"
    assert tea_tasting.utils.format_num(10_000_000) == "1.00e+07"
    assert tea_tasting.utils.format_num(12345, thousands_sep=" ") == "12 345"
    assert tea_tasting.utils.format_num(1.2345, decimal_point=",") == "1,23"
    assert tea_tasting.utils.format_num(0) == "0.00"


def test_get_and_format_num():
    data: dict[str, object] = {
        "name": "metric",
        "metric": 0.12345,
        "rel_metric": 0.12345,
        "metric_ci_lower": 0.12345,
        "metric_ci_upper": 0.98765,
        "rel_metric_ci_lower": 0.12345,
        "rel_metric_ci_upper": 0.98765,
        "power": 0.87654,
    }
    assert tea_tasting.utils.get_and_format_num(data, "name") == "metric"
    assert tea_tasting.utils.get_and_format_num(data, "metric") == "0.123"
    assert tea_tasting.utils.get_and_format_num(data, "rel_metric") == "12%"
    assert tea_tasting.utils.get_and_format_num(data, "metric_ci") == "[0.123, 0.988]"
    assert tea_tasting.utils.get_and_format_num(data, "rel_metric_ci") == "[12%, 99%]"
    assert tea_tasting.utils.get_and_format_num(data, "power") == "88%"


@pytest.fixture
def dicts_repr() -> tea_tasting.utils.DictsReprMixin:
    class DictsRepr(tea_tasting.utils.DictsReprMixin):
        default_keys = ("a", "b")
        def to_dicts(self) -> tuple[dict[str, object], ...]:
            return (
                {"a": 0.12345, "b": 0.23456},
                {"a": 0.34567, "b": 0.45678},
                {"a": 0.56789, "b": 0.67890},
            )
    return DictsRepr()

def test_dicts_repr_mixin_to_arrow(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr.to_arrow().equals(pa.table({
        "a": (0.12345, 0.34567, 0.56789),
        "b": (0.23456, 0.45678, 0.67890),
    }))

def test_dicts_repr_mixin_to_pandas(dicts_repr: tea_tasting.utils.DictsReprMixin):
    pandas.testing.assert_frame_equal(
        dicts_repr.to_pandas(),
        pd.DataFrame({
            "a": (0.12345, 0.34567, 0.56789),
            "b": (0.23456, 0.45678, 0.67890),
        }),
    )

def test_dicts_repr_mixin_to_polars(dicts_repr: tea_tasting.utils.DictsReprMixin):
    polars.testing.assert_frame_equal(
        dicts_repr.to_polars(),
        pl.DataFrame({
            "a": (0.12345, 0.34567, 0.56789),
            "b": (0.23456, 0.45678, 0.67890),
        }),
    )

def test_dicts_repr_mixin_to_pretty_dicts_default(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
):
    assert dicts_repr.to_pretty_dicts() == [
        {"a": "0.123", "b": "0.235"},
        {"a": "0.346", "b": "0.457"},
        {"a": "0.568", "b": "0.679"},
    ]

def test_dicts_repr_mixin_to_pretty_dicts_max_rows(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
):
    assert dicts_repr.to_pretty_dicts(max_rows=2) == [
        {"a": "0.123", "b": "0.235"},
        {"a": "…", "b": "…"},
        {"a": "0.568", "b": "0.679"},
    ]

def test_dicts_repr_mixin_to_string(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr.to_string() == textwrap.dedent("""\
            a     b
        0.123 0.235
        0.346 0.457
        0.568 0.679""")

def test_dicts_repr_mixin_to_html(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr.to_html() == (
        '<table class="dataframe" style="text-align: right;">'
        '<thead><tr><th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td>0.123</td><td>0.235</td></tr>'
        '<tr><td>0.346</td><td>0.457</td></tr>'
        '<tr><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_to_html_indent(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
):
    assert dicts_repr.to_html(indent="    ") == textwrap.dedent("""\
        <table class="dataframe" style="text-align: right;">
            <thead>
                <tr>
                    <th>a</th>
                    <th>b</th>
                </tr>
            </thead>
            <tbody>
                <tr>
                    <td>0.123</td>
                    <td>0.235</td>
                </tr>
                <tr>
                    <td>0.346</td>
                    <td>0.457</td>
                </tr>
                <tr>
                    <td>0.568</td>
                    <td>0.679</td>
                </tr>
            </tbody>
        </table>""")

def test_dicts_repr_mixin_with_defaults(dicts_repr: tea_tasting.utils.DictsReprMixin):
    new_dicts_repr =  dicts_repr.with_defaults(keys=("c", "d"), max_rows=42)
    assert new_dicts_repr.default_keys == ("c", "d")
    assert new_dicts_repr.default_max_rows == 42

def test_dicts_repr_mixin_with_keys(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr.with_keys(("c", "d")).default_keys == ("c", "d")

def test_dicts_repr_mixin_with_max_rows(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr.with_max_rows(42).default_max_rows == 42

def test_dicts_repr_mixin_mime_marimo(dicts_repr: tea_tasting.utils.DictsReprMixin):
    mime_type, data = dicts_repr._mime_()
    assert mime_type == "text/html"
    assert data.startswith("<marimo-ui-element")

def test_dicts_repr_mixin_mime_exception(dicts_repr: tea_tasting.utils.DictsReprMixin):
    def import_side_effect(name, *args, **kwargs):  # type: ignore  # noqa: ANN001, ANN002, ANN003, ANN202
        if name == "marimo":
            raise ImportError("No module named 'marimo'")
        return builtins.__import__(name, *args, **kwargs)
    with unittest.mock.patch("builtins.__import__", side_effect=import_side_effect):
        mime_type, data = dicts_repr._mime_()
    assert mime_type == "text/html"
    assert data.startswith('<table class="dataframe"')

def test_dicts_repr_mixin_repr_html(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert dicts_repr._repr_html_() == (
        '<table class="dataframe" style="text-align: right;">'
        '<thead><tr><th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td>0.123</td><td>0.235</td></tr>'
        '<tr><td>0.346</td><td>0.457</td></tr>'
        '<tr><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_repr(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert repr(dicts_repr) == textwrap.dedent("""\
            a     b
        0.123 0.235
        0.346 0.457
        0.568 0.679""")

def test_dicts_repr_mixin_str(dicts_repr: tea_tasting.utils.DictsReprMixin):
    assert str(dicts_repr) == textwrap.dedent("""\
            a     b
        0.123 0.235
        0.346 0.457
        0.568 0.679""")


def test_repr_mixin_repr():
    class Repr(tea_tasting.utils.ReprMixin):
        def __init__(self, a: int, *, b: bool, c: str) -> None:
            self._a = -1
            self.a_ = -1
            self.a = a
            self.b_ = -1
            self.b = b
            self.c = c
    r = Repr(a=1, b=False, c="c")
    assert repr(r) == f"Repr(a=1, b=False, c={'c'!r})"

def test_repr_mixin_repr_obj():
    class Obj(tea_tasting.utils.ReprMixin):
        ...
    obj = Obj()
    assert repr(obj) == "Obj()"

def test_repr_mixin_repr_pos():
    class Pos(tea_tasting.utils.ReprMixin):
        def __init__(self, *args: int) -> None:
            self.args = args
    pos = Pos(1, 2, 3)
    with pytest.raises(RuntimeError):
        repr(pos)


def test_div():
    assert tea_tasting.utils.div(1, 2) == 0.5
    assert tea_tasting.utils.div(1, 0, 3) == 3
    assert tea_tasting.utils.div(1, 0) == float("inf")
    assert math.isnan(tea_tasting.utils.div(0, 0))
    assert math.isnan(tea_tasting.utils.div(-1, 0))


def test_float():
    typ = tea_tasting.utils.Float
    assert typ(1) + 2 == typ(3)
    assert 1 + typ(2) == typ(3)
    assert typ(1) - 2 == typ(-1)
    assert 1 - typ(2) == typ(-1)
    assert typ(1) / 2 == 0.5
    assert 1 / typ(2) == 0.5
    assert math.isnan(typ(0) / 0)
    assert math.isnan(0 / typ(0))
    assert typ(1) / 0 == float("inf")
    assert 1 / typ(0) == float("inf")
    assert math.isnan(typ(-1) / 0)
    assert math.isnan(-1 / typ(0))
    assert typ(5) // 2 == typ(2)
    assert 5 // typ(2) == typ(2)
    assert typ(5) % 2 == typ(1)
    assert 5 % typ(2) == typ(1)
    assert divmod(typ(5), 2) == (typ(2), typ(1))
    assert divmod(5, typ(2)) == (typ(2), typ(1))
    assert typ(2) ** 3 == typ(8)
    assert 2 ** typ(3) == typ(8)
    assert -typ(1) == typ(-1)
    assert +typ(1) == typ(1)
    assert abs(typ(-1)) == typ(1)
    assert int(typ(1.0)) == 1
    assert float(typ(1.0)) == 1.0
    assert round(typ(11), -1) == typ(10)
    assert math.trunc(typ(1.2)) == typ(1)
    assert math.floor(typ(1.2)) == typ(1)
    assert math.ceil(typ(1.2)) == typ(2)


def test_int():
    typ = tea_tasting.utils.Int
    assert typ(1) + 2 == typ(3)
    assert 1 + typ(2) == typ(3)
    assert typ(1) - 2 == typ(-1)
    assert 1 - typ(2) == typ(-1)
    assert typ(1) / 2 == 0.5
    assert 1 / typ(2) == 0.5
    assert math.isnan(typ(0) / 0)
    assert math.isnan(0 / typ(0))
    assert typ(1) / 0 == float("inf")
    assert 1 / typ(0) == float("inf")
    assert math.isnan(typ(-1) / 0)
    assert math.isnan(-1 / typ(0))
    assert typ(5) // 2 == typ(2)
    assert 5 // typ(2) == typ(2)
    assert typ(5) % 2 == typ(1)
    assert 5 % typ(2) == typ(1)
    assert divmod(typ(5), 2) == (typ(2), typ(1))
    assert divmod(5, typ(2)) == (typ(2), typ(1))
    assert typ(2) ** 3 == typ(8)
    assert 2 ** typ(3) == typ(8)
    assert -typ(1) == typ(-1)
    assert +typ(1) == typ(1)
    assert abs(typ(-1)) == typ(1)
    assert int(typ(1.0)) == 1
    assert float(typ(1.0)) == 1.0
    assert round(typ(11), -1) == typ(10)
    assert math.trunc(typ(1.2)) == typ(1)
    assert math.floor(typ(1.2)) == typ(1)
    assert math.ceil(typ(1.2)) == typ(1)


def test_numeric():
    assert isinstance(tea_tasting.utils.numeric(1), tea_tasting.utils.Int)
    assert isinstance(tea_tasting.utils.numeric("1"), tea_tasting.utils.Int)
    assert isinstance(tea_tasting.utils.numeric(1.0), tea_tasting.utils.Float)
    assert isinstance(tea_tasting.utils.numeric("inf"), tea_tasting.utils.Float)



"""Module for working with aggregated statistics: count, mean, var, cov."""

from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, overload

import ibis.expr.operations
import ibis.expr.types
import narwhals as nw

import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Sequence

    import narwhals.typing  # noqa: TC004


_COUNT = "_count"
_MEAN = "_mean__{}"
_VAR = "_var__{}"
_COV = "_cov__{}__{}"
_DEMEAN = "_demean__{}"


class Aggregates(tea_tasting.utils.ReprMixin):  # noqa: D101
    count_: int | None
    mean_: dict[str, float | int]
    var_: dict[str, float | int]
    cov_: dict[tuple[str, str], float | int]

    def __init__(
        self,
        count_: int | None = None,
        mean_: dict[str, float | int] = {},  # noqa: B006
        var_: dict[str, float | int] = {},  # noqa: B006
        cov_: dict[tuple[str, str], float | int] = {},  # noqa: B006
    ) -> None:
        """Aggregated statistics.

        Args:
            count_: Sample size (number of observations).
            mean_: Dictionary of sample means with variable names as keys.
            var_: Dictionary of sample variances with variable names as keys.
            cov_: Dictionary of sample covariances with pairs of variable names as keys.
        """
        self.count_ = count_
        self.mean_ = mean_
        self.var_ = var_
        self.cov_ = {_sorted_tuple(*k): v for k, v in cov_.items()}

    def with_zero_div(self) -> Aggregates:
        """Return aggregates that do not raise an error on division by zero.

        Division by zero returns:

        - `inf` if numerator is greater than `0`,
        - `nan` if numerator is equal to or less than `0`.
        """
        return Aggregates(
            count_=None if self.count_ is None else tea_tasting.utils.Int(self.count_),
            mean_={k: tea_tasting.utils.numeric(v) for k, v in self.mean_.items()},
            var_={k: tea_tasting.utils.numeric(v) for k, v in self.var_.items()},
            cov_={k: tea_tasting.utils.numeric(v) for k, v in self.cov_.items()},
        )

    def count(self) -> int:
        """Sample size (number of observations).

        Returns:
            Sample size (number of observations).
        """
        if self.count_ is None:
            raise RuntimeError("Count is None.")
        return self.count_

    def mean(self, name: str | None) -> float | int:
        """Sample mean.

        Assume the variable is a constant `1` if the variable name is `None`.

        Args:
            name: Variable name.

        Returns:
            Sample mean.
        """
        if name is None:
            return 1
        return self.mean_[name]

    def var(self, name: str | None) -> float | int:
        """Sample variance.

        Assume the variable is a constant if the variable name is `None`.

        Args:
            name: Variable name.

        Returns:
            Sample variance.
        """
        if name is None:
            return 0
        return self.var_[name]

    def cov(self, left: str | None, right: str | None) -> float | int:
        """Sample covariance.

        Assume the variable is a constant if the variable name is `None`.

        Args:
            left: First variable name.
            right: Second variable name.

        Returns:
            Sample covariance.
        """
        if left is None or right is None:
            return 0
        return self.cov_[_sorted_tuple(left, right)]

    def ratio_var(
        self,
        numer: str | None,
        denom: str | None,
    ) -> float | int:
        """Sample variance of the ratio of two variables using the Delta method.

        Args:
            numer: Numerator variable name.
            denom: Denominator variable name.

        Returns:
            Sample variance of the ratio of two variables.

        References:
            - [Delta method](https://en.wikipedia.org/wiki/Delta_method).
            - [Taylor expansions for the moments of functions of random variables](https://en.wikipedia.org/wiki/Taylor_expansions_for_the_moments_of_functions_of_random_variables).
        """
        numer_mean_sq = self.mean(numer) * self.mean(numer)
        denom_mean_sq = self.mean(denom) * self.mean(denom)
        return (
            self.var(numer)
            - 2 * self.cov(numer, denom) * self.mean(numer) / self.mean(denom)
            + self.var(denom) * numer_mean_sq / denom_mean_sq
        ) / denom_mean_sq

    def ratio_cov(
        self,
        left_numer: str | None,
        left_denom: str | None,
        right_numer: str | None,
        right_denom: str | None,
    ) -> float | int:
        """Sample covariance of the ratios of variables using the Delta method.

        Args:
            left_numer: First numerator variable name.
            left_denom: First denominator variable name.
            right_numer: Second numerator variable name.
            right_denom: Second denominator variable name.

        Returns:
            Sample covariance of the ratios of variables.

        References:
            - [Delta method](https://en.wikipedia.org/wiki/Delta_method).
            - [Taylor expansions for the moments of functions of random variables](https://en.wikipedia.org/wiki/Taylor_expansions_for_the_moments_of_functions_of_random_variables).
        """
        left_ratio_of_means = self.mean(left_numer) / self.mean(left_denom)
        right_ratio_of_means = self.mean(right_numer) / self.mean(right_denom)
        return (
            self.cov(left_numer, right_numer)
            - self.cov(left_numer, right_denom) * right_ratio_of_means
            - self.cov(left_denom, right_numer) * left_ratio_of_means
            + self.cov(left_denom, right_denom)
                * left_ratio_of_means * right_ratio_of_means
        ) / self.mean(left_denom) / self.mean(right_denom)

    def __add__(self, other: Aggregates) -> Aggregates:
        """Calculate the aggregated statistics of the concatenation of two samples.

        Samples are assumed to be independent.

        Args:
            other: Aggregated statistics of the second sample.

        Returns:
            Aggregated statistics of the concatenation of two samples.
        """
        return Aggregates(
            count_=self.count() + other.count() if self.count_ is not None else None,
            mean_={col: _add_mean(self, other, col) for col in self.mean_},
            var_={col: _add_var(self, other, col) for col in self.var_},
            cov_={cols: _add_cov(self, other, cols) for cols in self.cov_},
        )


def _add_mean(left: Aggregates, right: Aggregates, col: str) -> float | int:
    sum_ = left.count()*left.mean(col) + right.count()*right.mean(col)
    count = left.count() + right.count()
    return sum_ / count

def _add_var(left: Aggregates, right: Aggregates, col: str) -> float | int:
    left_n = left.count()
    right_n = right.count()
    total_n = left_n + right_n
    diff_of_means = left.mean(col) - right.mean(col)
    return (
        left.var(col) * (left_n - 1)
        + right.var(col) * (right_n - 1)
        + diff_of_means * diff_of_means * left_n * right_n / total_n
    ) / (total_n - 1)

def _add_cov(left: Aggregates, right: Aggregates, cols: tuple[str, str]) -> float | int:
    left_n = left.count()
    right_n = right.count()
    total_n = left_n + right_n
    diff_of_means0 = left.mean(cols[0]) - right.mean(cols[0])
    diff_of_means1 = left.mean(cols[1]) - right.mean(cols[1])
    return (
        left.cov(*cols) * (left_n - 1)
        + right.cov(*cols) * (right_n - 1)
        + diff_of_means0 * diff_of_means1 * left_n * right_n / total_n
    ) / (total_n - 1)


@overload
def read_aggregates(
    data: ibis.expr.types.Table | narwhals.typing.IntoFrame,
    group_col: str,
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> dict[object, Aggregates]:
    ...

@overload
def read_aggregates(
    data: ibis.expr.types.Table | narwhals.typing.IntoFrame,
    group_col: None,
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> Aggregates:
    ...

def read_aggregates(
    data: ibis.expr.types.Table | narwhals.typing.IntoFrame,
    group_col: str | None,
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> dict[object, Aggregates] | Aggregates:
    """Extract aggregated statistics.

    Args:
        data: Granular data.
        group_col: Column name to group by before aggregation.
            If `None`, total aggregates are calculated.
        has_count: If `True`, calculate the sample size.
        mean_cols: Column names for calculation of sample means.
        var_cols: Column names for calculation of sample variances.
        cov_cols: Pairs of column names for calculation of sample covariances.

    Returns:
        Aggregated statistics.
    """
    mean_cols, var_cols, cov_cols = _validate_aggr_cols(mean_cols, var_cols, cov_cols)

    if isinstance(data, ibis.expr.types.Table):
        aggr_data = _read_aggr_ibis(
            data=data,
            group_col=group_col,
            has_count=has_count,
            mean_cols=mean_cols,
            var_cols=var_cols,
            cov_cols=cov_cols,
        )
    else:
        aggr_data = _read_aggr_narwhals(
            data=data,
            group_col=group_col,
            has_count=has_count,
            mean_cols=mean_cols,
            var_cols=var_cols,
            cov_cols=cov_cols,
        )

    if group_col is None:
        return _get_aggregates(
            aggr_data[0],
            has_count=has_count,
            mean_cols=mean_cols,
            var_cols=var_cols,
            cov_cols=cov_cols,
        )

    return {
        group_data[group_col]: _get_aggregates(
            group_data,
            has_count=has_count,
            mean_cols=mean_cols,
            var_cols=var_cols,
            cov_cols=cov_cols,
        )
        for group_data in aggr_data
    }


def _validate_aggr_cols(
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> tuple[tuple[str, ...], tuple[str, ...], tuple[tuple[str, str], ...]]:
    mean_cols = tuple({*mean_cols})
    var_cols = tuple({*var_cols})
    cov_cols = tuple({
        _sorted_tuple(left, right)
        for left, right in cov_cols
    })
    return mean_cols, var_cols, cov_cols


def _sorted_tuple(left: str, right: str) -> tuple[str, str]:
    if right < left:
        return right, left
    return left, right


def _read_aggr_ibis(
    data: ibis.expr.types.Table,
    group_col: str | None,
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> list[dict[str, int | float]]:
    covar_cols = tuple({*var_cols, *itertools.chain(*cov_cols)})
    backend = ibis.get_backend(data)
    var_op = ibis.expr.operations.Variance
    cov_op = ibis.expr.operations.Covariance
    if backend.has_operation(var_op) and backend.has_operation(cov_op):
        var_expr = {
            _VAR.format(col): data[col].cast("float").var(how="sample")  # type: ignore
            for col in var_cols
        }
        cov_expr = {
            _COV.format(left, right): data[left].cast("float").cov(  # type: ignore
                data[right].cast("float"),  # type: ignore
                how="sample",
            )
            for left, right in cov_cols
        }
    else:
        # Use demeaned values if backend doesn't have var and cov functions.
        if len(covar_cols) > 0:
            demean_expr = {
                _DEMEAN.format(col): data[col] - data[col].cast("float").mean()  # type: ignore
                for col in covar_cols
            }
            grouped_data = data.group_by(group_col) if group_col is not None else data  # type: ignore
            data = grouped_data.mutate(**demean_expr)  # type: ignore

        var_expr = {
            _VAR.format(col): (
                data[_DEMEAN.format(col)] * data[_DEMEAN.format(col)]
            ).sum() / (data.count() - 1)  # type: ignore
            for col in var_cols
        }
        cov_expr = {
            _COV.format(left, right): (
                data[_DEMEAN.format(left)] * data[_DEMEAN.format(right)]
            ).sum() / (data.count() - 1)  # type: ignore
            for left, right in cov_cols
        }

    count_expr = {_COUNT: data.count()} if has_count else {}
    mean_expr = {_MEAN.format(col): data[col].cast("float").mean() for col in mean_cols}  # type: ignore
    all_expr = count_expr | mean_expr | var_expr | cov_expr

    grouped_data = data.group_by(group_col) if group_col is not None else data  # type: ignore
    return grouped_data.aggregate(**all_expr).to_pyarrow().to_pylist()  # type: ignore


def _read_aggr_narwhals(
    data: narwhals.typing.IntoFrame,
    group_col: str | None,
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> list[dict[str, int | float]]:
    data = nw.from_native(data)
    if not isinstance(data, nw.LazyFrame):
        data = data.lazy()

    covar_cols = tuple({*var_cols, *itertools.chain(*cov_cols)})
    if len(covar_cols) > 0:
        data = (
            data.with_columns(**{
                _DEMEAN.format(col): _demean_nw_col(col, group_col)
                for col in covar_cols
            })
            .with_columns(
                **{
                    _VAR.format(col):
                        nw.col(_DEMEAN.format(col)) * nw.col(_DEMEAN.format(col))
                    for col in var_cols
                },
                **{
                    _COV.format(left, right):
                        nw.col(_DEMEAN.format(left)) * nw.col(_DEMEAN.format(right))
                    for left, right in cov_cols
                },
            )
        )

    count_expr = {_COUNT: nw.len()} if has_count or len(covar_cols) > 0 else {}
    mean_expr = {_MEAN.format(col): nw.col(col).mean() for col in mean_cols}
    var_expr = {_VAR.format(col): nw.col(_VAR.format(col)).mean() for col in var_cols}
    cov_expr = {
        _COV.format(left, right): nw.col(_COV.format(left, right)).mean()
        for left, right in cov_cols
    }
    all_expr = count_expr | mean_expr | var_expr | cov_expr

    aggr_data = (
        data.select(**all_expr) if group_col is None
        else data.group_by(group_col).agg(**all_expr)
    )
    if len(covar_cols) > 0:
        aggr_data = aggr_data.with_columns(
            **{
                _VAR.format(col): nw.col(_VAR.format(col)) / (1 - 1/nw.col(_COUNT))
                for col in var_cols
            },
            **{
                _COV.format(left, right): nw.col(_COV.format(left, right)) /
                    (1 - 1/nw.col(_COUNT))
                for left, right in cov_cols
            },
        )

    return aggr_data.collect().to_arrow().to_pylist()


def _demean_nw_col(col: str, group_col: str | None) -> nw.Expr:
    if group_col is None:
        return nw.col(col) - nw.col(col).mean()
    return nw.col(col) - nw.col(col).mean().over(group_col)


def _get_aggregates(
    data: dict[str, float | int],
    *,
    has_count: bool,
    mean_cols: Sequence[str],
    var_cols: Sequence[str],
    cov_cols: Sequence[tuple[str, str]],
) -> Aggregates:
    return Aggregates(
        count_=data[_COUNT] if has_count else None,  # type: ignore
        mean_={col: data[_MEAN.format(col)] for col in mean_cols},
        var_={col: data[_VAR.format(col)] for col in var_cols},
        cov_={cols: data[_COV.format(*cols)] for cols in cov_cols},
    )



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple
import unittest.mock

import pytest

import tea_tasting.aggr
import tea_tasting.datasets
import tea_tasting.metrics.base
import tea_tasting.metrics.proportion


if TYPE_CHECKING:
    import pyarrow as pa


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, seed=42)

@pytest.fixture
def data_aggr(data_arrow: pa.Table) -> dict[object, tea_tasting.aggr.Aggregates]:
    return tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col="variant",
        has_count=True,
        mean_cols=(),
        var_cols=(),
        cov_cols=(),
    )


def test_sample_ratio_init_default():
    metric = tea_tasting.metrics.proportion.SampleRatio()
    assert metric.ratio == 1
    assert metric.method == "auto"
    assert metric.correction is True

def test_sample_ratio_init_custom():
    metric = tea_tasting.metrics.proportion.SampleRatio(
        {0: 0.5, 1: 0.5},
        method="norm",
        correction=False,
    )
    assert metric.ratio == {0: 0.5, 1: 0.5}
    assert metric.method == "norm"
    assert metric.correction is False


def test_sample_ratio_aggr_cols():
    metric = tea_tasting.metrics.proportion.SampleRatio()
    assert metric.aggr_cols == tea_tasting.metrics.base.AggrCols(has_count=True)


def test_sample_ratio_analyze_frame(data_arrow: pa.Table):
    metric = tea_tasting.metrics.proportion.SampleRatio()
    result = metric.analyze(data_arrow, 0, 1, variant="variant")
    assert isinstance(result, tea_tasting.metrics.proportion.SampleRatioResult)

def test_sample_ratio_analyze_auto():
    metric = tea_tasting.metrics.proportion.SampleRatio()
    with unittest.mock.patch("scipy.stats.binomtest") as mock:
        mock.return_value = NamedTuple("Result", (("pvalue", float),))(pvalue=0.1)
        data = tea_tasting.datasets.make_users_data(
            seed=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD - 1,
        )
        metric.analyze(data, 0, 1, variant="variant")
        mock.assert_called_once()
    with unittest.mock.patch("scipy.stats.norm.sf") as mock:
        mock.return_value = 0.1
        data = tea_tasting.datasets.make_users_data(
            seed=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD,
        )
        metric.analyze(data, 0, 1, variant="variant")
        mock.assert_called_once()

def test_sample_ratio_analyze_binom(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.proportion.SampleRatio(method="binom")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.6172994135892521)

def test_sample_ratio_analyze_norm_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.proportion.SampleRatio(method="norm", correction=True)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.6170750774519738)

def test_sample_ratio_analyze_norm_no_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.proportion.SampleRatio(method="norm", correction=False)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.5485062355001472)

def test_sample_ratio_analyze_aggregates(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
):
    metric = tea_tasting.metrics.proportion.SampleRatio()
    with pytest.raises(NotImplementedError):
        metric.analyze_aggregates(data_aggr[0], data_aggr[1])



"""Example datasets."""
# ruff: noqa: PLR0913

from __future__ import annotations

from typing import TYPE_CHECKING, overload

import numpy as np
import pyarrow as pa

import tea_tasting.utils


if TYPE_CHECKING:
    from typing import Literal

    import numpy.typing as npt
    import pandas as pd
    import polars as pl


@overload
def make_users_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["arrow"] = "arrow",
) -> pa.Table:
    ...

@overload
def make_users_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["pandas"],
) -> pd.DataFrame:
    ...

@overload
def make_users_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["polars"],
) -> pl.DataFrame:
    ...

def make_users_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["arrow", "pandas", "polars"] = "arrow",
) -> pa.Table | pd.DataFrame | pl.DataFrame:
    """Generate simulated data for A/B testing scenarios.

    Data mimics what you might encounter in an A/B test for an online store,
    with a user-level randomization. Each row represents an individual user
    with information about:

    - `user`: User identifier.
    - `variant`: Variant of the test. 0 is control, 1 is treatment.
    - `sessions`: Number of user's sessions.
    - `orders`: Number of user's orders.
    - `revenue`: Revenue generated by the user.

    Optionally, pre-experimental data can be generated as well:

    - `sessions_covariate`: Number of user's sessions
        before the experiment.
    - `orders_covariate`: Number of user's orders before the experiment.
    - `revenue_covariate`: Revenue generated by the user
        before the experiment.

    Args:
        covariates: If `True`, generates pre-experimental data as the covariates
            in addition to default columns.
        seed: Random seed.
        n_users: Number of users.
        ratio: Ratio of the number of users in treatment relative to control.
        sessions_uplift: Sessions uplift in the treatment variant, relative to control.
        orders_uplift: Orders uplift in the treatment variant, relative to control.
        revenue_uplift: Revenue uplift in the treatment variant, relative to control.
        avg_sessions: Average number of sessions per user.
        avg_orders_per_session: Average number of orders per session.
            Should be less than `1`.
        avg_revenue_per_order: Average revenue per order.
        return_type: Return type:

            - `"arrow"`: PyArrow Table.
            - `"pandas"`: Pandas DataFrame.
            - `"polars"`: Polars DataFrame.

    Returns:
        Simulated data for A/B testing scenarios.

    Examples:
        ```pycon
        >>> import tea_tasting as tt

        >>> data = tt.make_users_data(seed=42)
        >>> data
        pyarrow.Table
        user: int64
        variant: int64
        sessions: int64
        orders: int64
        revenue: double
        ----
        user: [[0,1,2,3,4,...,3995,3996,3997,3998,3999]]
        variant: [[1,0,1,1,0,...,0,0,0,0,0]]
        sessions: [[2,2,2,2,1,...,2,2,3,1,5]]
        orders: [[1,1,1,1,1,...,0,0,0,0,2]]
        revenue: [[9.17,6.43,7.94,15.93,7.14,...,0,0,0,0,17.16]]

        ```

        With covariates:

        ```pycon
        >>> data = tt.make_users_data(seed=42, covariates=True)
        >>> data
        pyarrow.Table
        user: int64
        variant: int64
        sessions: int64
        orders: int64
        revenue: double
        sessions_covariate: int64
        orders_covariate: int64
        revenue_covariate: double
        ----
        user: [[0,1,2,3,4,...,3995,3996,3997,3998,3999]]
        variant: [[1,0,1,1,0,...,0,0,0,0,0]]
        sessions: [[2,2,2,2,1,...,2,2,3,1,5]]
        orders: [[1,1,1,1,1,...,0,0,0,0,2]]
        revenue: [[9.17,6.43,7.94,15.93,7.14,...,0,0,0,0,17.16]]
        sessions_covariate: [[3,4,4,1,1,...,1,3,2,1,5]]
        orders_covariate: [[2,1,2,0,1,...,0,1,0,0,0]]
        revenue_covariate: [[19.19,2.77,22.57,0,13.68,...,0,13.52,0,0,0]]

        ```

        As Pandas DataFrame:

        ```pycon
        >>> data = tt.make_users_data(seed=42, return_type="pandas")
        >>> data
              user  variant  sessions  orders  revenue
        0        0        1         2       1     9.17
        1        1        0         2       1     6.43
        2        2        1         2       1     7.94
        3        3        1         2       1    15.93
        4        4        0         1       1     7.14
        ...    ...      ...       ...     ...      ...
        3995  3995        0         2       0     0.00
        3996  3996        0         2       0     0.00
        3997  3997        0         3       0     0.00
        3998  3998        0         1       0     0.00
        3999  3999        0         5       2    17.16
        <BLANKLINE>
        [4000 rows x 5 columns]

        ```

        As Polars DataFrame:

        ```pycon
        >>> data = tt.make_users_data(seed=42, return_type="polars")
        >>> data
        shape: (4_000, 5)
        ┌──────┬─────────┬──────────┬────────┬─────────┐
        │ user ┆ variant ┆ sessions ┆ orders ┆ revenue │
        │ ---  ┆ ---     ┆ ---      ┆ ---    ┆ ---     │
        │ i64  ┆ i64     ┆ i64      ┆ i64    ┆ f64     │
        ╞══════╪═════════╪══════════╪════════╪═════════╡
        │ 0    ┆ 1       ┆ 2        ┆ 1      ┆ 9.17    │
        │ 1    ┆ 0       ┆ 2        ┆ 1      ┆ 6.43    │
        │ 2    ┆ 1       ┆ 2        ┆ 1      ┆ 7.94    │
        │ 3    ┆ 1       ┆ 2        ┆ 1      ┆ 15.93   │
        │ 4    ┆ 0       ┆ 1        ┆ 1      ┆ 7.14    │
        │ …    ┆ …       ┆ …        ┆ …      ┆ …       │
        │ 3995 ┆ 0       ┆ 2        ┆ 0      ┆ 0.0     │
        │ 3996 ┆ 0       ┆ 2        ┆ 0      ┆ 0.0     │
        │ 3997 ┆ 0       ┆ 3        ┆ 0      ┆ 0.0     │
        │ 3998 ┆ 0       ┆ 1        ┆ 0      ┆ 0.0     │
        │ 3999 ┆ 0       ┆ 5        ┆ 2      ┆ 17.16   │
        └──────┴─────────┴──────────┴────────┴─────────┘

        ```
    """
    return _make_data(
        covariates=covariates,
        seed=seed,
        n_users=n_users,
        ratio=ratio,
        sessions_uplift=sessions_uplift,
        orders_uplift=orders_uplift,
        revenue_uplift=revenue_uplift,
        avg_sessions=avg_sessions,
        avg_orders_per_session=avg_orders_per_session,
        avg_revenue_per_order=avg_revenue_per_order,
        return_type=return_type,
        explode_sessions=False,
    )


@overload
def make_sessions_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["arrow"] = "arrow",
) -> pa.Table:
    ...

@overload
def make_sessions_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["pandas"],
) -> pd.DataFrame:
    ...

@overload
def make_sessions_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["polars"],
) -> pl.DataFrame:
    ...

def make_sessions_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["arrow", "pandas", "polars"] = "arrow",
) -> pa.Table | pd.DataFrame | pl.DataFrame:
    """Generate simulated user data for A/B testing scenarios.

    Data mimics what you might encounter in an A/B test for an online store,
    with a user-level randomization. Each row represents a user's session
    with information about:

    - `user`: User identifier.
    - `variant`: Variant of the test. 0 is control, 1 is treatment.
    - `sessions`: Number of user's sessions.
    - `orders`: Number of user's orders.
    - `revenue`: Revenue generated by the user.

    Optionally, pre-experimental data can be generated as well:

    - `sessions_covariate`: Number of user's sessions
        before the experiment.
    - `orders_covariate`: Number of user's orders before the experiment.
    - `revenue_covariate`: Revenue generated by the user
        before the experiment.

    Args:
        covariates: If `True`, generates pre-experimental data as the covariates
            in addition to default columns.
        seed: Random seed.
        n_users: Number of users.
        ratio: Ratio of the number of users in treatment relative to control.
        sessions_uplift: Sessions uplift in the treatment variant, relative to control.
        orders_uplift: Orders uplift in the treatment variant, relative to control.
        revenue_uplift: Revenue uplift in the treatment variant, relative to control.
        avg_sessions: Average number of sessions per user.
        avg_orders_per_session: Average number of orders per session.
            Should be less than `1`.
        avg_revenue_per_order: Average revenue per order.
        return_type: Return type:

            - `"arrow"`: PyArrow Table.
            - `"pandas"`: Pandas DataFrame.
            - `"polars"`: Polars DataFrame.

    Returns:
        Simulated data for A/B testing scenarios.

    Examples:
        ```pycon
        >>> import tea_tasting as tt

        >>> data = tt.make_sessions_data(seed=42)
        >>> data
        pyarrow.Table
        user: int64
        variant: int64
        sessions: int64
        orders: int64
        revenue: double
        ----
        user: [[0,0,1,1,2,...,3999,3999,3999,3999,3999]]
        variant: [[1,1,0,0,1,...,0,0,0,0,0]]
        sessions: [[1,1,1,1,1,...,1,1,1,1,1]]
        orders: [[1,1,1,1,1,...,1,0,1,1,0]]
        revenue: [[5.89,6.13,2.61,12.3,11.57,...,23.63,0,2.4,24.54,0]]

        ```

        With covariates:

        ```pycon
        >>> data = tt.make_sessions_data(seed=42, covariates=True)
        >>> data
        pyarrow.Table
        user: int64
        variant: int64
        sessions: int64
        orders: int64
        revenue: double
        sessions_covariate: double
        orders_covariate: double
        revenue_covariate: double
        ----
        user: [[0,0,1,1,2,...,3999,3999,3999,3999,3999]]
        variant: [[1,1,0,0,1,...,0,0,0,0,0]]
        sessions: [[1,1,1,1,1,...,1,1,1,1,1]]
        orders: [[1,1,1,1,1,...,1,0,1,1,0]]
        revenue: [[5.89,6.13,2.61,12.3,11.57,...,23.63,0,2.4,24.54,0]]
        sessions_covariate: [[1.5,1.5,0,0,1.5,...,0.2,0.2,0.2,0.2,0.2]]
        orders_covariate: [[0.5,0.5,0,0,1.5,...,0,0,0,0,0]]
        revenue_covariate: [[1.24,1.24,0,0,12.32,...,0,0,0,0,0]]

        ```

        As Pandas DataFrame:

        ```pycon
        >>> data = tt.make_sessions_data(seed=42, return_type="pandas")
        >>> data
              user  variant  sessions  orders  revenue
        0        0        1         1       1     5.89
        1        0        1         1       1     6.13
        2        1        0         1       1     2.61
        3        1        0         1       1    12.30
        4        2        1         1       1    11.57
        ...    ...      ...       ...     ...      ...
        7953  3999        0         1       1    23.63
        7954  3999        0         1       0     0.00
        7955  3999        0         1       1     2.40
        7956  3999        0         1       1    24.54
        7957  3999        0         1       0     0.00
        <BLANKLINE>
        [7958 rows x 5 columns]

        ```

        As Polars DataFrame:

        ```pycon
        >>> data = tt.make_sessions_data(seed=42, return_type="polars")
        >>> data
        shape: (7_958, 5)
        ┌──────┬─────────┬──────────┬────────┬─────────┐
        │ user ┆ variant ┆ sessions ┆ orders ┆ revenue │
        │ ---  ┆ ---     ┆ ---      ┆ ---    ┆ ---     │
        │ i64  ┆ i64     ┆ i64      ┆ i64    ┆ f64     │
        ╞══════╪═════════╪══════════╪════════╪═════════╡
        │ 0    ┆ 1       ┆ 1        ┆ 1      ┆ 5.89    │
        │ 0    ┆ 1       ┆ 1        ┆ 1      ┆ 6.13    │
        │ 1    ┆ 0       ┆ 1        ┆ 1      ┆ 2.61    │
        │ 1    ┆ 0       ┆ 1        ┆ 1      ┆ 12.3    │
        │ 2    ┆ 1       ┆ 1        ┆ 1      ┆ 11.57   │
        │ …    ┆ …       ┆ …        ┆ …      ┆ …       │
        │ 3999 ┆ 0       ┆ 1        ┆ 1      ┆ 23.63   │
        │ 3999 ┆ 0       ┆ 1        ┆ 0      ┆ 0.0     │
        │ 3999 ┆ 0       ┆ 1        ┆ 1      ┆ 2.4     │
        │ 3999 ┆ 0       ┆ 1        ┆ 1      ┆ 24.54   │
        │ 3999 ┆ 0       ┆ 1        ┆ 0      ┆ 0.0     │
        └──────┴─────────┴──────────┴────────┴─────────┘

        ```
    """
    return _make_data(
        covariates=covariates,
        seed=seed,
        n_users=n_users,
        ratio=ratio,
        sessions_uplift=sessions_uplift,
        orders_uplift=orders_uplift,
        revenue_uplift=revenue_uplift,
        avg_sessions=avg_sessions,
        avg_orders_per_session=avg_orders_per_session,
        avg_revenue_per_order=avg_revenue_per_order,
        return_type=return_type,
        explode_sessions=True,
    )


def _make_data(
    *,
    covariates: bool = False,
    seed: int | np.random.Generator | np.random.SeedSequence | None = None,
    n_users: int = 4000,
    ratio: float | int = 1,
    sessions_uplift: float | int = 0.0,
    orders_uplift: float = 0.1,
    revenue_uplift: float = 0.1,
    avg_sessions: float | int = 2,
    avg_orders_per_session: float = 0.25,
    avg_revenue_per_order: float | int = 10,
    return_type: Literal["arrow", "pandas", "polars"] = "arrow",
    explode_sessions: bool = False,
) -> pa.Table | pd.DataFrame | pl.DataFrame:
    _check_params(
        n_users=n_users,
        ratio=ratio,
        sessions_uplift=sessions_uplift,
        orders_uplift=orders_uplift,
        revenue_uplift=revenue_uplift,
        avg_sessions=avg_sessions,
        avg_orders_per_session=avg_orders_per_session,
        avg_revenue_per_order=avg_revenue_per_order,
    )

    rng = np.random.default_rng(seed=seed)
    user = np.arange(n_users)
    variant = rng.binomial(n=1, p=ratio / (1 + ratio), size=n_users)
    sessions_mult = 1 + sessions_uplift*variant
    sessions = 1 + rng.poisson(lam=avg_sessions*sessions_mult - 1, size=n_users)
    size = n_users
    orders_per_sessions_sample_size = 1  # Parameter of Beta distribution (alpha+beta).
    revenue_log_scale = 0.5  # Parameter of log-normal distribution.

    if explode_sessions:
        user = np.repeat(user, sessions)
        sessions = np.ones_like(user)
        size = len(user)
        revenue_log_scale = np.sqrt(np.log(
            1 + avg_sessions*(np.exp(revenue_log_scale**2) - 1)))

    orders_per_sessions_mult = (1 + orders_uplift*variant) / (
        1 + sessions_uplift*variant)
    orders_per_sessions = rng.beta(
        a=avg_orders_per_session * orders_per_sessions_mult
            * orders_per_sessions_sample_size,
        b=(1 - avg_orders_per_session*orders_per_sessions_mult)
            * orders_per_sessions_sample_size,
        size=n_users,
    )
    orders = rng.binomial(n=sessions, p=orders_per_sessions[user], size=size)

    revenue_per_order_mult = (1 + revenue_uplift*variant) / (1 + orders_uplift*variant)
    revenue_per_order = rng.lognormal(
        mean=(
            np.log(avg_revenue_per_order * revenue_per_order_mult[user])
            - revenue_log_scale * revenue_log_scale / 2
        ),
        sigma=revenue_log_scale,
        size=size,
    )

    revenue = orders * revenue_per_order

    data = {
        "user": user,
        "variant": variant[user],
        "sessions": sessions,
        "orders": orders,
        "revenue": revenue.round(2),
    }

    if covariates:
        sessions_covariate = rng.poisson(lam=sessions / sessions_mult[user], size=size)
        orders_per_sessions_covariate = orders_per_sessions / orders_per_sessions_mult

        orders_covariate = rng.binomial(
            n=sessions_covariate,
            p=orders_per_sessions_covariate[user],
            size=size,
        )

        revenue_per_order_covariate = rng.lognormal(
            mean=(
                np.log(revenue_per_order / revenue_per_order_mult[user])
                - revenue_log_scale * revenue_log_scale / 2
            ),
            sigma=revenue_log_scale,
            size=size,
        )

        revenue_covariate = orders_covariate * revenue_per_order_covariate

        if explode_sessions:
            sessions_covariate = _avg_by_groups(sessions_covariate, user)
            orders_covariate = _avg_by_groups(orders_covariate, user)
            revenue_covariate = _avg_by_groups(revenue_covariate, user)

        data |= {
            "sessions_covariate": sessions_covariate,
            "orders_covariate": orders_covariate,
            "revenue_covariate": revenue_covariate.round(2),
        }

    if return_type == "pandas":
        import pandas as pd
        return pd.DataFrame(data)
    if return_type == "polars":
        import polars as pl
        return pl.DataFrame(data)
    return pa.table(data)


def _check_params(
    *,
    n_users: int,
    ratio: float | int,
    sessions_uplift: float | int,
    orders_uplift: float,
    revenue_uplift: float,
    avg_sessions: float | int,
    avg_orders_per_session: float,
    avg_revenue_per_order: float | int,
) -> None:
    tea_tasting.utils.check_scalar(n_users, name="n_users", typ=int, ge=10)
    tea_tasting.utils.check_scalar(ratio, name="ratio", typ=float | int, gt=0)
    tea_tasting.utils.check_scalar(
        sessions_uplift, name="sessions_uplift", typ=float | int, gt=1/avg_sessions - 1)
    tea_tasting.utils.check_scalar(
        orders_uplift,
        name="orders_uplift",
        typ=float | int,
        gt=-1,
        lt=(1 + sessions_uplift)/avg_orders_per_session - 1,
    )
    tea_tasting.utils.check_scalar(
        revenue_uplift, name="revenue_uplift", typ=float | int, gt=-1)
    tea_tasting.utils.check_scalar(
        avg_sessions, name="avg_sessions", typ=float | int, gt=1)
    tea_tasting.utils.check_scalar(
        avg_orders_per_session, name="avg_orders_per_session", typ=float, gt=0, lt=1)
    tea_tasting.utils.check_scalar(
        avg_revenue_per_order, name="avg_revenue_per_order", typ=float | int, gt=0)


def _avg_by_groups(
    values: npt.NDArray[np.number],
    groups: npt.NDArray[np.number],
) -> npt.NDArray[np.number]:
    return np.concatenate([
        np.full(v.shape, v.mean())
        for v in np.split(values, np.unique(groups, return_index=True)[1])
        if len(v) > 0
    ])



"""Global configuration."""
# ruff: noqa: PLR0913

from __future__ import annotations

import contextlib
import contextvars
from typing import TYPE_CHECKING, overload

import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Literal


_DEFAULT_CONFIG: dict[str, object] = {
    "alpha": 0.05,
    "alternative": "two-sided",
    "confidence_level": 0.95,
    "equal_var": False,
    "n_obs": None,
    "n_resamples": 10_000,
    "power": 0.8,
    "ratio": 1,
    "use_t": True,
}

_config_var: contextvars.ContextVar[dict[str, object]] = contextvars.ContextVar(
    "tea_tasting.config",
    default=_DEFAULT_CONFIG.copy(),  # noqa: B039
)


@overload
def get_config(option: Literal["alpha"]) -> float:
    ...

@overload
def get_config(option: Literal["alternative"]) -> str:
    ...

@overload
def get_config(option: Literal["confidence_level"]) -> float:
    ...

@overload
def get_config(option: Literal["equal_var"]) -> bool:
    ...

@overload
def get_config(option: Literal["n_obs"]) -> int | Sequence[int] | None:
    ...

@overload
def get_config(option: Literal["n_resamples"]) -> str:
    ...

@overload
def get_config(option: Literal["power"]) -> float:
    ...

@overload
def get_config(option: Literal["ratio"]) -> float | int:
    ...

@overload
def get_config(option: Literal["use_t"]) -> bool:
    ...

@overload
def get_config(option: str) -> object:
    ...

@overload
def get_config(option: None = None) -> dict[str, object]:
    ...

def get_config(option: str | None = None) -> object:
    """Retrieve the current settings of the global configuration.

    Args:
        option: The option name.

    Returns:
        The specified option value if its name is provided,
            or a dictionary containing all options otherwise.

    Examples:
        ```pycon
        >>> import tea_tasting as tt

        >>> tt.get_config("equal_var")
        False

        ```
    """
    config = _config_var.get()
    return config[option] if option is not None else config.copy()


def _set_config(**params: object) -> contextvars.Token[dict[str, object]]:
    config = _config_var.get().copy()
    for name, value in params.items():
        if value is not None:
            config[name] = tea_tasting.utils.auto_check(value, name)
    return _config_var.set(config)


def set_config(
    *,
    alpha: float | None = None,
    alternative: Literal["two-sided", "greater", "less"] | None = None,
    confidence_level: float | None = None,
    equal_var: bool | None = None,
    n_obs: int | Sequence[int] | None = None,
    n_resamples: int | None = None,
    power: float | None = None,
    ratio: float | int | None = None,
    use_t: bool | None = None,
    **kwargs: object,
) -> None:
    """Update the global configuration with specified settings.

    Args:
        alpha: Significance level. Default is 0.05.
        alternative: Alternative hypothesis:

            - `"two-sided"`: the means are unequal,
            - `"greater"`: the mean in the treatment variant is greater than the mean
                in the control variant,
            - `"less"`: the mean in the treatment variant is less than the mean
                in the control variant.

            Default is `"two-sided"`.

        confidence_level: Confidence level for the confidence interval.
            Default is `0.95`.
        equal_var: Defines whether equal variance is assumed. If `True`,
            pooled variance is used for the calculation of the standard error
            of the difference between two means. Default is `False`.
        n_obs: Number of observations in the control and in the treatment together.
            Default is `None`.
        n_resamples: The number of resamples performed to form the bootstrap
            distribution of a statistic. Default is `10_000`.
        power: Statistical power. Default is 0.8.
        ratio: Ratio of the number of observations in the treatment
            relative to the control. Default is 1.
        use_t: Defines whether to use the Student's t-distribution (`True`) or
            the Normal distribution (`False`) by default. Default is `True`.
        **kwargs: User-defined global parameters.

    Examples:
        ```pycon
        >>> import tea_tasting as tt

        >>> tt.set_config(equal_var=True, use_t=False)
        >>> experiment = tt.Experiment(
        ...     sessions_per_user=tt.Mean("sessions"),
        ...     orders_per_session=tt.RatioOfMeans("orders", "sessions"),
        ...     orders_per_user=tt.Mean("orders"),
        ...     revenue_per_user=tt.Mean("revenue"),
        ... )
        >>> tt.set_config(equal_var=False, use_t=True)
        >>> experiment.metrics["orders_per_user"]
        Mean(value='orders', covariate=None, alternative='two-sided', confidence_level=0.95, equal_var=True, use_t=False, alpha=0.05, ratio=1, power=0.8, effect_size=None, rel_effect_size=None, n_obs=None)

        ```
    """  # noqa: E501
    _set_config(**{k: v for k, v in locals().items() if k != "kwargs"}, **kwargs)


@contextlib.contextmanager
def config_context(
    *,
    alpha: float | None = None,
    alternative: Literal["two-sided", "greater", "less"] | None = None,
    confidence_level: float | None = None,
    equal_var: bool | None = None,
    n_obs: int | Sequence[int] | None = None,
    n_resamples: int | None = None,
    power: float | None = None,
    ratio: float | int | None = None,
    use_t: bool | None = None,
    **kwargs: object,
) -> Iterator[object]:
    """A context manager that temporarily modifies the global configuration.

    Args:
        alpha: Significance level. Default is 0.05.
        alternative: Alternative hypothesis:

            - `"two-sided"`: the means are unequal,
            - `"greater"`: the mean in the treatment variant is greater than the mean
                in the control variant,
            - `"less"`: the mean in the treatment variant is less than the mean
                in the control variant.

            Default is `"two-sided"`.

        confidence_level: Confidence level for the confidence interval.
            Default is `0.95`.
        equal_var: Defines whether equal variance is assumed. If `True`,
            pooled variance is used for the calculation of the standard error
            of the difference between two means. Default is `False`.
        n_obs: Number of observations in the control and in the treatment together.
            Default is `None`.
        n_resamples: The number of resamples performed to form the bootstrap
            distribution of a statistic. Default is `10_000`.
        power: Statistical power. Default is 0.8.
        ratio: Ratio of the number of observations in the treatment
            relative to the control. Default is 1.
        use_t: Defines whether to use the Student's t-distribution (`True`) or
            the Normal distribution (`False`) by default. Default is `True`.
        **kwargs: User-defined global parameters.

    Examples:
        ```pycon
        >>> import tea_tasting as tt

        >>> with tt.config_context(equal_var=True, use_t=False):
        ...     experiment = tt.Experiment(
        ...         sessions_per_user=tt.Mean("sessions"),
        ...         orders_per_session=tt.RatioOfMeans("orders", "sessions"),
        ...         orders_per_user=tt.Mean("orders"),
        ...         revenue_per_user=tt.Mean("revenue"),
        ...     )
        >>> experiment.metrics["orders_per_user"]
        Mean(value='orders', covariate=None, alternative='two-sided', confidence_level=0.95, equal_var=True, use_t=False, alpha=0.05, ratio=1, power=0.8, effect_size=None, rel_effect_size=None, n_obs=None)

        ```
    """  # noqa: E501
    token = _set_config(
        **{k: v for k, v in locals().items() if k != "kwargs"},
        **kwargs,
    )
    try:
        yield
    finally:
        _config_var.reset(token)



"""Package version."""

from __future__ import annotations

import importlib.metadata
import importlib.resources


try:
    __version__ = importlib.metadata.version(__package__ or "tea-tasting")
except importlib.metadata.PackageNotFoundError:
    __version__ = (
        importlib.resources.files("tea_tasting")
        .joinpath("_version.txt")
        .read_text()
        .strip()
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest

import tea_tasting.config
import tea_tasting.datasets
import tea_tasting.metrics.base
import tea_tasting.metrics.resampling


if TYPE_CHECKING:
    import numpy.typing as npt
    import pyarrow as pa


@pytest.fixture
def data_arrow() -> pa.Table:
    return tea_tasting.datasets.make_users_data(n_users=100, seed=42)

@pytest.fixture
def data_gran(data_arrow: pa.Table) -> dict[object, pa.Table]:
    return tea_tasting.metrics.base.read_granular(
        data_arrow,
        ("sessions", "orders", "revenue"),
        variant="variant",
    )


def test_bootstrap_init_default():
    metric = tea_tasting.metrics.resampling.Bootstrap("a", np.mean)
    assert metric.columns == "a"
    assert metric.statistic == np.mean
    assert metric.alternative == tea_tasting.config.get_config("alternative")
    assert metric.confidence_level == tea_tasting.config.get_config("confidence_level")
    assert metric.n_resamples == tea_tasting.config.get_config("n_resamples")
    assert metric.method == "bca"
    assert metric.batch is None
    assert metric.random_state is None

def test_bootstrap_init_custom():
    metric = tea_tasting.metrics.resampling.Bootstrap(
        ("a", "b"),
        np.mean,
        alternative="greater",
        confidence_level=0.9,
        n_resamples=1000,
        method="basic",
        batch=100,
        random_state=42,
    )
    assert metric.columns == ("a", "b")
    assert metric.statistic == np.mean
    assert metric.alternative == "greater"
    assert metric.confidence_level == 0.9
    assert metric.n_resamples == 1000
    assert metric.method == "basic"
    assert metric.batch == 100
    assert metric.random_state == 42


def test_bootstrap_cols():
    metric = tea_tasting.metrics.resampling.Bootstrap("a", np.mean)
    assert metric.cols == ("a",)

    metric = tea_tasting.metrics.resampling.Bootstrap(("a", "b"), np.mean)
    assert metric.cols == ("a", "b")


def test_bootstrap_analyze_frame(data_arrow: pa.Table):
    metric = tea_tasting.metrics.resampling.Bootstrap("sessions", np.mean)
    result = metric.analyze(data_arrow, 0, 1, variant="variant")
    assert isinstance(result, tea_tasting.metrics.resampling.BootstrapResult)


def test_bootstrap_analyze_default(data_gran: dict[object, pa.Table]):
    metric = tea_tasting.metrics.resampling.Bootstrap(
        "revenue",
        np.mean,
        n_resamples=100,
        random_state=42,
    )
    result = metric.analyze(data_gran, 0, 1)
    assert isinstance(result, tea_tasting.metrics.resampling.BootstrapResult)
    assert result.control == pytest.approx(5.029811320754717)
    assert result.treatment == pytest.approx(5.43)
    assert result.effect_size == pytest.approx(0.4001886792452831)
    assert result.effect_size_ci_lower == pytest.approx(-3.269396309565539)
    assert result.effect_size_ci_upper == pytest.approx(7.219843380442667)
    assert result.rel_effect_size == pytest.approx(0.07956335809137971)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.5658493834599828)
    assert result.rel_effect_size_ci_upper == pytest.approx(1.8185473860534842)

def test_bootstrap_analyze_multiple_columns(data_gran: dict[object, pa.Table]):
    def ratio_of_means(
        sample: npt.NDArray[np.number],
        axis: int,
    ) -> npt.NDArray[np.number]:
        stat = np.mean(sample, axis=axis)  # type: ignore
        return stat[0] / stat[1]

    metric = tea_tasting.metrics.resampling.Bootstrap(
        ("orders", "sessions"),
        ratio_of_means,
        n_resamples=100,
        random_state=42,
    )
    result = metric.analyze(data_gran, 0, 1)
    assert isinstance(result, tea_tasting.metrics.resampling.BootstrapResult)
    assert result.control == pytest.approx(0.2857142857142857)
    assert result.treatment == pytest.approx(0.20224719101123595)
    assert result.effect_size == pytest.approx(-0.08346709470304975)
    assert result.effect_size_ci_lower == pytest.approx(-0.24780839493679777)
    assert result.effect_size_ci_upper == pytest.approx(0.07730723504025493)
    assert result.rel_effect_size == pytest.approx(-0.2921348314606741)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.6424902672606227)
    assert result.rel_effect_size_ci_upper == pytest.approx(0.4374404130492657)

def test_bootstrap_analyze_division_by_zero(data_gran: dict[object, pa.Table]):
    metric = tea_tasting.metrics.resampling.Bootstrap(
        "orders",
        np.median,
        n_resamples=100,
        random_state=42,
        method="basic",
    )
    result = metric.analyze(data_gran, 0, 1)
    assert isinstance(result, tea_tasting.metrics.resampling.BootstrapResult)
    assert result.control == 0
    assert result.treatment == 0
    assert result.effect_size == 0
    assert result.effect_size_ci_lower == 0
    assert result.effect_size_ci_upper == 0
    assert np.isnan(result.rel_effect_size)
    assert np.isnan(result.rel_effect_size_ci_lower)
    assert np.isnan(result.rel_effect_size_ci_upper)

def test_quantile(data_gran: dict[object, pa.Table]):
    metric = tea_tasting.metrics.resampling.Quantile(
        "revenue",
        q=0.8,
        alternative="greater",
        confidence_level=0.9,
        n_resamples=100,
        random_state=42,
    )
    assert metric.column == "revenue"
    assert metric.q == 0.8
    result = metric.analyze(data_gran, 0, 1)
    assert isinstance(result, tea_tasting.metrics.resampling.BootstrapResult)
    assert result.control == pytest.approx(11.972000000000001)
    assert result.treatment == pytest.approx(6.2820000000000045)
    assert result.effect_size == pytest.approx(-5.689999999999997)
    assert result.effect_size_ci_lower == pytest.approx(-10.875800000000003)
    assert result.effect_size_ci_upper == float("inf")
    assert result.rel_effect_size == pytest.approx(-0.47527564316739024)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.8743329817472134)
    assert result.rel_effect_size_ci_upper == float("inf")



"""Useful functions and classes."""
# ruff: noqa: SIM114
# pyright: reportOperatorIssue=false

from __future__ import annotations

import abc
from collections.abc import Sequence
import inspect
import locale
import math
from typing import TYPE_CHECKING, overload
import xml.etree.ElementTree as ET

import pyarrow as pa


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Literal, TypeAlias, TypeVar

    import pandas as pd
    import polars as pl


    DictsReprMixinT = TypeVar("DictsReprMixinT", bound="DictsReprMixin")
    R = TypeVar("R")


def check_scalar(  # noqa: PLR0913
    value: R,
    name: str = "value",
    *,
    typ: object = None,
    ge: object = None,
    gt: object = None,
    le: object = None,
    lt: object = None,
    ne: object = None,
    in_: object = None,
) -> R:
    """Check if a scalar parameter meets specified type and value constraints.

    Args:
        value: Parameter value.
        name: Parameter name.
        typ: Acceptable data types.
        ge: If not `None`, check that parameter value is greater than
            or equal to `ge`.
        gt: If not `None`, check that parameter value is greater than `gt`.
        le: If not `None`, check that parameter value is less than or equal to `le`.
        lt: If not `None`, check that parameter value is less than `lt`.
        ne: If not `None`, check that parameter value is not equal to `ne`.
        in_: If not `None`, check that parameter value is in `in_`.

    Returns:
        Parameter value.
    """
    if typ is not None and not isinstance(value, typ):  # type: ignore
        raise TypeError(f"{name} must be an instance of {typ}.")
    if ge is not None and value < ge:
        raise ValueError(f"{name} == {value}, must be >= {ge}.")
    if gt is not None and value <= gt:
        raise ValueError(f"{name} == {value}, must be > {gt}.")
    if le is not None and value > le:
        raise ValueError(f"{name} == {value}, must be <= {le}.")
    if lt is not None and value >= lt:
        raise ValueError(f"{name} == {value}, must be < {lt}.")
    if ne is not None and value == ne:
        raise ValueError(f"{name} == {value}, must be != {ne}.")
    if in_ is not None and value not in in_:
        raise ValueError(f"{name} == {value}, must be in {in_}.")
    return value


@overload
def auto_check(value: float, name: Literal["alpha"]) -> float:
    ...

@overload
def auto_check(value: str, name: Literal["alternative"]) -> str:
    ...

@overload
def auto_check(value: float, name: Literal["confidence_level"]) -> float:
    ...

@overload
def auto_check(value: bool, name: Literal["correction"]) -> bool:  # noqa: FBT001
    ...

@overload
def auto_check(value: bool, name: Literal["equal_var"]) -> bool:  # noqa: FBT001
    ...

@overload
def auto_check(
    value: int | Sequence[int] | None,
    name: Literal["n_obs"],
) -> int | Sequence[int] | None:
    ...

@overload
def auto_check(value: int, name: Literal["n_resamples"]) -> int:
    ...

@overload
def auto_check(value: float, name: Literal["power"]) -> float:
    ...

@overload
def auto_check(value: float | int, name: Literal["ratio"]) -> float | int:
    ...

@overload
def auto_check(value: bool, name: Literal["use_t"]) -> bool:  # noqa: FBT001
    ...

@overload
def auto_check(value: R, name: str) -> R:
    ...

def auto_check(value: R, name: str) -> R:  # noqa: C901, PLR0912
    """Automatically check a parameter's type and value based on its name.

    The following parameter names are supported: `"alpha"`, `"alternative"`,
    `"confidence_level"`, `"correction"`, `"equal_var"`, `"n_obs"`,
    `"n_resamples"`, `"power"`, `"ratio"`, `"use_t"`.

    Args:
        value: Parameter value.
        name: Parameter name.

    Returns:
        Parameter value.
    """
    if name == "alpha":
        check_scalar(value, name, typ=float, gt=0, lt=1)
    if name == "alternative":
        check_scalar(value, name, typ=str, in_={"two-sided", "greater", "less"})
    elif name == "confidence_level":
        check_scalar(value, name, typ=float, gt=0, lt=1)
    elif name == "correction":
        check_scalar(value, name, typ=bool)
    elif name == "equal_var":
        check_scalar(value, name, typ=bool)
    elif name == "n_obs":
        check_scalar(value, name, typ=int | Sequence | None)
        if isinstance(value, int):
            check_scalar(value, name, gt=1)
        if isinstance(value, Sequence):
            for val in value:
                check_scalar(val, name, typ=int, gt=1)
    elif name == "n_resamples":
        check_scalar(value, name, typ=int, gt=0)
    elif name == "power":
        check_scalar(value, name, typ=float, gt=0, lt=1)
    elif name == "ratio":
        check_scalar(value, name, typ=float | int, gt=0)
    elif name == "use_t":
        check_scalar(value, name, typ=bool)
    return value


def format_num(
    val: float | int | None,
    sig: int = 3,
    *,
    pct: bool = False,
    nan: str = "-",
    inf: str = "∞",
    fixed_point_range: tuple[float | None, float | None] = (0.001, 10_000_000),
    thousands_sep: str | None = None,
    decimal_point: str | None = None,
) -> str:
    """Format a number according to specified formatting rules.

    Args:
        val: Number to format.
        sig: Number of significant digits.
        pct: If `True`, format as a percentage.
        nan: Replacement for `None` and `nan` values.
        inf: Replacement for infinite values.
        fixed_point_range: The range within which the number is formatted
            as fixed point.
            Number outside of the range is formatted as exponential.
            `None` means no boundary.
        thousands_sep: Thousands separator. If `None`, the value from locales is used.
        decimal_point: Decimal point symbol. If `None`, the value from locales is used.

    Returns:
        Formatted number.
    """
    if val is None or math.isnan(val):
        return nan

    if math.isinf(val):
        return inf if val > 0 else "-" + inf

    if pct:
        val = val * 100

    if (
        (fixed_point_range[0] is not None and abs(val) < fixed_point_range[0]) or
        (fixed_point_range[1] is not None and abs(val) >= fixed_point_range[1])
    ):
        precision = max(0, sig - 1)
        typ = "e" if val != 0 else "f"
    else:
        precision = max(0, sig - 1 - math.floor(math.log10(abs(val))))
        val = round(val, precision)
        # Repeat in order to format 99.999 as "100", not "100.0".
        precision = max(0, sig - 1 - math.floor(math.log10(abs(val))))
        typ = "f"

    result = format(val, f"_.{precision}{typ}")

    if thousands_sep is None:
        thousands_sep = locale.localeconv().get("thousands_sep", "_")  # type: ignore
    if thousands_sep != "_":
        result = result.replace("_", thousands_sep)

    if decimal_point is None:
        decimal_point = locale.localeconv().get("decimal_point", ".")  # type: ignore
    if decimal_point != ".":
        result = result.replace(".", decimal_point)

    if pct:
        return result + "%"

    return result


def get_and_format_num(data: dict[str, object], key: str) -> str:
    """Get and format dictionary value.

    Formatting rules:

    - If a name starts with `"rel_"` or equals to `"power"` consider it
        a percentage value. Round percentage values to 2 significant digits,
        multiply by `100` and add `"%"`.
    - Round other values to 3 significant values.
    - If value is less than `0.001` or is greater than or equal to `10_000_000`,
        format it in exponential presentation.
    - If a name ends with `"_ci"`, consider it a confidence interval.
        Look up for attributes `"{name}_lower"` and `"{name}_upper"`,
        and format the interval as `"[{lower_bound}, {upper_bound}]"`.

    Args:
        data: Dictionary.
        key: Key.

    Returns:
        Formatted value.
    """
    if key.endswith("_ci"):
        ci_lower = get_and_format_num(data, key + "_lower")
        ci_upper = get_and_format_num(data, key + "_upper")
        return f"[{ci_lower}, {ci_upper}]"

    val = data.get(key)
    if not isinstance(val, float | int | None):
        return str(val)

    sig, pct = (2, True) if key.startswith("rel_") or key == "power" else (3, False)
    return format_num(val, sig=sig, pct=pct)


def _cache_method(
    method: Callable[[DictsReprMixinT], R],
) -> Callable[[DictsReprMixinT], R]:
    def cached_method(self: DictsReprMixinT) -> R:
        if self._cache is None:
            self._cache = {}
        if method.__name__ not in self._cache:
            self._cache[method.__name__] = method(self)
        return self._cache[method.__name__]  # type: ignore
    return cached_method

class DictsReprMixin(abc.ABC):
    """Representation and conversion of a sequence of dictionaries.

    Default formatting rules:

    - If a name starts with `"rel_"` or equals to `"power"` consider it
        a percentage value. Round percentage values to 2 significant digits,
        multiply by `100` and add `"%"`.
    - Round other values to 3 significant values.
    - If value is less than `0.001` or is greater than or equal to `10_000_000`,
        format it in exponential presentation.
    - If a name ends with `"_ci"`, consider it a confidence interval.
        Look up for attributes `"{name}_lower"` and `"{name}_upper"`,
        and format the interval as `"[{lower_bound}, {upper_bound}]"`.
    """
    default_keys: Sequence[str]
    default_max_rows: int = 0
    _cache: dict[str, object] | None = None

    @abc.abstractmethod
    def to_dicts(self) -> Sequence[dict[str, object]]:
        """Convert the object to a sequence of dictionaries."""

    @_cache_method
    def to_arrow(self) -> pa.Table:
        """Convert the object to a PyArrow Table."""
        return pa.Table.from_pylist(self.to_dicts())

    @_cache_method
    def to_pandas(self) -> pd.DataFrame:
        """Convert the object to a Pandas DataFrame."""
        import pandas as pd
        return pd.DataFrame.from_records(self.to_dicts())

    @_cache_method
    def to_polars(self) -> pl.DataFrame:
        """Convert the object to a Polars DataFrame."""
        import polars as pl
        return pl.from_dicts(self.to_dicts())


    def to_pretty_dicts(
        self,
        keys: Sequence[str] | None = None,
        formatter: Callable[[dict[str, object], str], str] = get_and_format_num,
        *,
        max_rows: int | None = None,
    ) -> list[dict[str, str]]:
        """Convert the object to a list of dictionaries with formatted values.

        Default formatting rules:

        - If a name starts with `"rel_"` or equals to `"power"` consider it
            a percentage value. Round percentage values to 2 significant digits,
            multiply by `100` and add `"%"`.
        - Round other values to 3 significant values.
        - If value is less than `0.001` or is greater than or equal to `10_000_000`,
            format it in exponential presentation.
        - If a name ends with `"_ci"`, consider it a confidence interval.
            Look up for attributes `"{name}_lower"` and `"{name}_upper"`,
            and format the interval as `"[{lower_bound}, {upper_bound}]"`.

        Args:
            keys: Keys to convert. If a key is not defined in the dictionary
                it's assumed to be `None`.
            formatter: Custom formatter function. It should accept a dictionary
                of metric result attributes and an attribute name, and return
                a formatted attribute value.
            max_rows: Maximum number of rows to convert.
                If `None`, the default value will be used.
                If `0` or less, all rows will be converted.

        Returns:
            List of dictionaries with formatted values.
        """
        if keys is None:
            keys = self.default_keys
        if max_rows is None:
            max_rows = self.default_max_rows

        dicts = self.to_dicts()
        if max_rows <= 0 or len(dicts) <= max_rows:
            return [{key: formatter(data, key) for key in keys} for data in dicts]

        bottom = max_rows // 2
        top = max_rows - bottom
        return (
            [{key: formatter(data, key) for key in keys} for data in dicts[:top]] +
            [dict.fromkeys(keys, "…")] +
            [{key: formatter(data, key) for key in keys} for data in dicts[-bottom:]]
        )


    def to_string(
        self,
        keys: Sequence[str] | None = None,
        formatter: Callable[[dict[str, object], str], str] = get_and_format_num,
        *,
        max_rows: int | None = None,
    ) -> str:
        """Convert the object to a string.

        Default formatting rules:

        - If a name starts with `"rel_"` or equals to `"power"` consider it
            a percentage value. Round percentage values to 2 significant digits,
            multiply by `100` and add `"%"`.
        - Round other values to 3 significant values.
        - If value is less than `0.001` or is greater than or equal to `10_000_000`,
            format it in exponential presentation.
        - If a name ends with `"_ci"`, consider it a confidence interval.
            Look up for attributes `"{name}_lower"` and `"{name}_upper"`,
            and format the interval as `"[{lower_bound}, {upper_bound}]"`.

        Args:
            keys: Keys to convert. If a key is not defined in the dictionary
                it's assumed to be `None`.
            formatter: Custom formatter function. It should accept a dictionary
                of metric result attributes and an attribute name, and return
                a formatted attribute value.
            max_rows: Maximum number of rows to convert.
                If `None`, the default value will be used.
                If `0` or less, all rows will be converted.

        Returns:
            A table with results rendered as string.
        """
        if keys is None:
            keys = self.default_keys
        if max_rows is None:
            max_rows = self.default_max_rows

        pretty_dicts = self.to_pretty_dicts(keys, formatter, max_rows=max_rows)
        widths = {key: len(key) for key in keys}
        for pretty_dict in pretty_dicts:
            for key in keys:
                widths[key] = max(widths[key], len(pretty_dict[key]))

        sep = " "
        rows = [sep.join(key.rjust(widths[key]) for key in keys)]
        rows.extend(
            sep.join(pretty_dict[key].rjust(widths[key]) for key in keys)
            for pretty_dict in pretty_dicts
        )
        return "\n".join(rows)


    def to_html(
        self,
        keys: Sequence[str] | None = None,
        formatter: Callable[[dict[str, object], str], str] = get_and_format_num,
        *,
        max_rows: int | None = None,
        indent: str | None = None,
    ) -> str:
        """Convert the object to HTML.

        Default formatting rules:

        - If a name starts with `"rel_"` or equals to `"power"` consider it
            a percentage value. Round percentage values to 2 significant digits,
            multiply by `100` and add `"%"`.
        - Round other values to 3 significant values.
        - If value is less than `0.001` or is greater than or equal to `10_000_000`,
            format it in exponential presentation.
        - If a name ends with `"_ci"`, consider it a confidence interval.
            Look up for attributes `"{name}_lower"` and `"{name}_upper"`,
            and format the interval as `"[{lower_bound}, {upper_bound}]"`.

        Args:
            keys: Keys to convert. If a key is not defined in the dictionary
                it's assumed to be `None`.
            formatter: Custom formatter function. It should accept a dictionary
                of metric result attributes and an attribute name, and return
                a formatted attribute value.
            max_rows: Maximum number of rows to convert.
                If `None`, the default value will be used.
                If `0` or less, all rows will be converted.
            indent: Whitespace to insert for each indentation level. If `None`,
                do not indent.

        Returns:
            A table with results rendered as HTML.
        """
        if keys is None:
            keys = self.default_keys
        if max_rows is None:
            max_rows = self.default_max_rows

        table = ET.Element(
            "table",
            {"class": "dataframe", "style": "text-align: right;"},
        )
        thead = ET.SubElement(table, "thead")
        thead_tr = ET.SubElement(thead, "tr")
        for key in keys:
            th = ET.SubElement(thead_tr, "th")
            th.text = key
        tbody = ET.SubElement(table, "tbody")
        for pretty_dict in self.to_pretty_dicts(keys, formatter, max_rows=max_rows):
            tr = ET.SubElement(tbody, "tr")
            for key in keys:
                td = ET.SubElement(tr, "td")
                td.text = pretty_dict[key]
        if indent is not None:
            ET.indent(table, space=indent)
        return ET.tostring(table, encoding="unicode", method="html")


    def with_defaults(
        self: DictsReprMixinT,
        *,
        keys: Sequence[str] | None = None,
        max_rows: int | None = None,
    ) -> DictsReprMixinT:
        """Copies the object and sets the new default parameters.

        Args:
            keys: New default `keys` for the methods `to_pretty_dicts`, `to_string`,
                and `to_html`.
            max_rows: New default `max_rows` for the methods `to_pretty_dicts`,
                `to_string`, and `to_html`.

        Returns:
            A copy of the object with the new default keys.
        """
        new_instance = self.__class__.__new__(self.__class__)
        new_instance.__dict__.update(self.__dict__)
        new_instance._cache = None
        if keys is not None:
            new_instance.default_keys = keys
        if max_rows is not None:
            new_instance.default_max_rows = max_rows
        return new_instance


    def with_keys(self: DictsReprMixinT, keys: Sequence[str]) -> DictsReprMixinT:
        """Copies the object and sets the new default `keys`.

        Args:
            keys: New default `keys` for the methods `to_pretty_dicts`, `to_string`,
                and `to_html`.

        Returns:
            A copy of the object with the new default `keys`.
        """
        return self.with_defaults(keys=keys)


    def with_max_rows(self: DictsReprMixinT, max_rows: int) -> DictsReprMixinT:
        """Copies the object and sets the new default `max_rows`.

        Args:
            max_rows: New default `max_rows` for the methods `to_pretty_dicts`,
                `to_string`, and `to_html`.

        Returns:
            A copy of the object with the new default `max_rows`.
        """
        return self.with_defaults(max_rows=max_rows)


    def _mime_(self) -> tuple[str, str]:
        """"Object representation for marimo notebooks."""
        try:
            import marimo as mo

            pretty_dicts = self.to_pretty_dicts(max_rows=0)
            n_rows = len(pretty_dicts)
            return mo.ui.table(  # type: ignore
                pretty_dicts,  # type: ignore
                selection=None,
                pagination=self.default_max_rows > 0 and n_rows > self.default_max_rows,
                page_size=self.default_max_rows
                    if self.default_max_rows > 0 and n_rows > self.default_max_rows
                    else None,
            )._mime_()
        except Exception:  # noqa: BLE001
            return "text/html", self._repr_html_()


    @_cache_method
    def _repr_html_(self) -> str:
        """Object HTML representation."""
        return self.to_html()

    @_cache_method
    def __repr__(self) -> str:
        """Object representation."""
        return self.__str__()

    @_cache_method
    def __str__(self) -> str:
        """Object string representation."""
        return self.to_string()


class ReprMixin:
    """A mixin class that provides a method for generating a string representation.

    Representation string is generated based on parameters values saved in attributes.
    """
    @classmethod
    def _get_param_names(cls) -> Iterator[str]:
        if cls.__init__ is object.__init__:
            return
        init_signature = inspect.signature(cls.__init__)

        for p in init_signature.parameters.values():
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "There should not be positional parameters in the __init__.")
            if p.name != "self" and p.kind != p.VAR_KEYWORD:
                yield p.name

    def __repr__(self) -> str:
        """Object representation."""
        params = {p: getattr(self, p) for p in self._get_param_names()}
        params_repr = ", ".join(f"{k}={v!r}" for k, v in params.items())
        return f"{self.__class__.__name__}({params_repr})"


def div(
    numer: float | int,
    denom: float | int,
    fill_zero_div: float | int | Literal["auto"] = "auto",
) -> float |int:
    """Perform division, providing specified results for cases of division by zero.

    Args:
        numer: Numerator.
        denom: Denominator.
        fill_zero_div: Result if denominator is zero.

    Returns:
        Result of the division.

    If `fill_zero_div` is equal `"auto"`, return:

    - `inf` if numerator is greater than `0`,
    - `nan` if numerator is equal to or less than `0`.
    """
    if denom != 0:
        return numer / denom
    if fill_zero_div != "auto":
        return fill_zero_div
    return float("inf") if numer > 0 else float("nan")


class _NumericBase:
    value: float | int
    fill_zero_div: float | int | Literal["auto"] = "auto"

    def __add__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(x + y, self.fill_zero_div)

    def __sub__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(x - y, self.fill_zero_div)

    def __mul__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(x * y, self.fill_zero_div)

    def __truediv__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(div(x, y, self.fill_zero_div), self.fill_zero_div)  # type: ignore

    def __floordiv__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(x // y, self.fill_zero_div)

    def __mod__(self, other: object) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        return numeric(x % y, self.fill_zero_div)

    def __divmod__(self, other: object) -> tuple[Numeric, Numeric]:
        x, y = self.value, getattr(other, "value", other)
        d, m = divmod(x, y)  # type: ignore
        return numeric(d, self.fill_zero_div), numeric(m, self.fill_zero_div)

    def __pow__(self, other: object, mod: object = None) -> Numeric:
        x, y = self.value, getattr(other, "value", other)
        z = getattr(mod, "value", mod)
        return numeric(pow(x, y, z), self.fill_zero_div)  # type: ignore

    def __radd__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(x + y, self.fill_zero_div)

    def __rsub__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(x - y, self.fill_zero_div)

    def __rmul__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(x * y, self.fill_zero_div)

    def __rtruediv__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(div(x, y, self.fill_zero_div), self.fill_zero_div)  # type: ignore

    def __rfloordiv__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(x // y, self.fill_zero_div)

    def __rmod__(self, other: object) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        return numeric(x % y, self.fill_zero_div)

    def __rdivmod__(self, other: object) -> tuple[Numeric, Numeric]:
        y, x = self.value, getattr(other, "value", other)
        d, m = divmod(x, y)  # type: ignore
        return numeric(d, self.fill_zero_div), numeric(m, self.fill_zero_div)

    def __rpow__(self, other: object, mod: object = None) -> Numeric:
        y, x = self.value, getattr(other, "value", other)
        z = getattr(mod, "value", mod)
        return numeric(pow(x, y, z), self.fill_zero_div)  # type: ignore

    def __neg__(self) -> Numeric:
        return numeric(-self.value, self.fill_zero_div)

    def __pos__(self) -> Numeric:
        return numeric(self)

    def __abs__(self) -> Numeric:
        return numeric(abs(self.value), self.fill_zero_div)

    def __float__(self) -> float:
        return float(self.value)

    def __int__(self) -> int:
        return int(self.value)

    def __round__(self, ndigits: int | None = None) -> Numeric:
        return numeric(round(self.value, ndigits), self.fill_zero_div)

    def __trunc__(self) -> Numeric:
        return numeric(math.trunc(self.value), self.fill_zero_div)

    def __floor__(self) -> Numeric:
        return numeric(math.floor(self.value), self.fill_zero_div)

    def __ceil__(self) -> Numeric:
        return numeric(math.ceil(self.value), self.fill_zero_div)


class Float(_NumericBase, float):
    """Float that gracefully handles division by zero errors."""
    def __new__(
        cls,
        value: object,
        fill_zero_div: float | int | Literal["auto"] = "auto",
    ) -> Float:
        """Float that gracefully handles division by zero errors."""
        instance = float.__new__(cls, value)  # type: ignore
        instance.value = float(value)  # type: ignore
        instance.fill_zero_div = fill_zero_div
        return instance

class Int(_NumericBase, int):
    """Integer that gracefully handles division by zero errors."""
    def __new__(
        cls,
        value: object,
        fill_zero_div: float | int | Literal["auto"] = "auto",
    ) -> Int:
        """Integer that gracefully handles division by zero errors."""
        instance = int.__new__(cls, value)  # type: ignore
        instance.value = int(value)  # type: ignore
        instance.fill_zero_div = fill_zero_div
        return instance

Numeric: TypeAlias = Float | Int


def numeric(
    value: object,
    fill_zero_div: float | int | Literal["auto"] = "auto",
) -> Numeric:
    """Convert an object to numeric that gracefully handles division by zero errors.

    Args:
        value: Object to convert.
        fill_zero_div: Result if denominator is zero.

    If `fill_zero_div` is equal `"auto"`, division by zero will return:

    - `inf` if numerator is greater than `0`,
    - `nan` if numerator is equal to or less than `0`.

    Returns:
        Float or integer that gracefully handles division by zero errors.
    """
    if isinstance(value, int):
        return Int(value, fill_zero_div)
    if isinstance(value, float):
        return Float(value, fill_zero_div)
    try:
        return Int(value, fill_zero_div)
    except ValueError:
        return Float(value, fill_zero_div)



from __future__ import annotations

import importlib
import importlib.metadata
import unittest.mock

import tea_tasting.version


def test_version():
    assert isinstance(tea_tasting.version.__version__, str)

    with (
        unittest.mock.patch(
            "tea_tasting.version.importlib.metadata.version") as version,
        unittest.mock.patch("tea_tasting.version.importlib.resources.files") as files,
    ):
        (
            files.return_value
            .joinpath.return_value
            .read_text.return_value
            .strip.return_value
        ) = "version"

        version.side_effect = importlib.metadata.PackageNotFoundError("Not found")
        importlib.reload(tea_tasting.version)
        assert isinstance(tea_tasting.version.__version__, str)



"""Metrics for the analysis of proportions."""

from __future__ import annotations

import math
from typing import TYPE_CHECKING, NamedTuple

import scipy.stats

import tea_tasting.aggr
import tea_tasting.metrics
from tea_tasting.metrics.base import AggrCols, MetricBaseAggregated
import tea_tasting.utils


if TYPE_CHECKING:
    from typing import Literal

    import ibis.expr.types
    import narwhals.typing


_MAX_EXACT_THRESHOLD = 1000


class SampleRatioResult(NamedTuple):
    """Result of the sample ratio mismatch check.

    Attributes:
        control: Number of observations in control.
        treatment: Number of observations in treatment.
        pvalue: P-value
    """
    control: float
    treatment: float
    pvalue: float


class SampleRatio(MetricBaseAggregated[SampleRatioResult]):  # noqa: D101
    def __init__(
        self,
        ratio: float | int | dict[object, float | int] = 1,
        *,
        method: Literal["auto", "binom", "norm"] = "auto",
        correction: bool = True,
    ) -> None:
        """Metric for sample ratio mismatch check.

        Args:
            ratio: Expected ratio of the number of observations in the treatment
                relative to the control.
            method: Statistical test used for calculation of p-value:

                - `"auto"`: Apply exact binomial test if the total number
                    of observations is < 1000; or normal approximation otherwise.
                - `"binom"`: Apply exact binomial test.
                - `"norm"`: Apply normal approximation of the binomial distribution.

            correction: If `True`, add continuity correction.
                Only for normal approximation.

        Examples:
            ```pycon
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     sample_ratio=tt.SampleRatio(),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result.with_keys(("metric", "control", "treatment", "pvalue"))
                  metric control treatment pvalue
            sample_ratio    2023      1977  0.477

            ```

            Different expected ratio:

            ```pycon
            >>> experiment = tt.Experiment(
            ...     sample_ratio=tt.SampleRatio(0.5),
            ... )
            >>> data = tt.make_users_data(seed=42)
            >>> result = experiment.analyze(data)
            >>> result.with_keys(("metric", "control", "treatment", "pvalue"))
                  metric control treatment    pvalue
            sample_ratio    2023      1977 3.26e-103

            ```
        """
        if isinstance(ratio, dict):
            for val in ratio.values():
                tea_tasting.utils.auto_check(val, "ratio")
        else:
            tea_tasting.utils.auto_check(ratio, "ratio")
        self.ratio = ratio

        self.method = tea_tasting.utils.check_scalar(
            method, "method", typ=str, in_={"auto", "binom", "norm"})
        self.correction = tea_tasting.utils.auto_check(correction, "correction")


    @property
    def aggr_cols(self) -> AggrCols:
        """Columns to be aggregated for a metric analysis."""
        return AggrCols(has_count=True)


    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table | dict[
            object, tea_tasting.aggr.Aggregates],
        control: object,
        treatment: object,
        variant: str | None = None,
    ) -> SampleRatioResult:
        """Perform a sample ratio mismatch check.

        Args:
            data: Experimental data.
            control: Control variant.
            treatment: Treatment variant.
            variant: Variant column name.

        Returns:
            Analysis result.
        """
        tea_tasting.utils.check_scalar(variant, "variant", typ=str | None)
        aggr = tea_tasting.metrics.aggregate_by_variants(
            data,
            aggr_cols=self.aggr_cols,
            variant=variant,
        )

        k = aggr[treatment].count()
        n = k + aggr[control].count()

        r = (
            self.ratio
            if isinstance(self.ratio, float | int)
            else self.ratio[treatment] / self.ratio[control]
        )
        p = r / (1 + r)

        if (
            self.method == "binom" or
            (self.method == "auto" and n < _MAX_EXACT_THRESHOLD)
        ):
            pvalue = scipy.stats.binomtest(k=int(k), n=int(n), p=p).pvalue
        else:  # norm
            d = k - n*p
            if self.correction and d != 0:
                d = min(d + 0.5, 0) if d < 0 else max(d - 0.5, 0)
            z = d / math.sqrt(n * p * (1 - p))
            pvalue = 2 * scipy.stats.norm.sf(abs(z))

        return SampleRatioResult(
            control=n - k,
            treatment=k,
            pvalue=pvalue,  # type: ignore
        )


    def analyze_aggregates(
        self,
        control: tea_tasting.aggr.Aggregates,
        treatment: tea_tasting.aggr.Aggregates,
    ) -> SampleRatioResult:
        """Stub method for compatibility with the base class."""
        raise NotImplementedError
